import numpy as np
import numpy.random

import arithmetics
import lrsNash
import tree
from tree import ptr
from matplotlib import pyplot as plt
import pickle

np.set_printoptions(precision=4, suppress=True)


class GoofSpiel(object):
    def __init__(self):
        self.M = 3
        self.N = 3
        self.upcard = [1, 1, 1, 1]
        self.valueA = [5.00, 1.33, 2.71, 1.80]
        self.valueB = [4.12, 6.28, 3.33, 1.92]
        self.T = tree.Tree()
        self.root = None

    def tree_init(self):
        self.root = self.T.new_node()
        self.root.parent = ptr(0)
        self.root.just_happened = np.array([0, 0])
        self.root.actA = np.arange(self.M) + 1
        self.root.actB = np.arange(self.N) + 1
        ### self.root.x = arithmetics.rand_f(1, len(self.root.actA) - 1)[0]
        ### self.root.y = arithmetics.rand_f(1, len(self.root.actB) - 1)[0]

        bfs_layers = []
        parent_layer = [+self.root]
        child_layer = []
        while True:
            bfs_layers.append(parent_layer)
            for p in parent_layer:
                n = self.T.MEM[p]
                n.child = np.resize(n.child, [len(n.actA), len(n.actB)])

                for i in range(len(n.actA)):
                    for j in range(len(n.actB)):
                        child = self.T.new_node()
                        n.child[i, j] = +child
                        child.parent = +n
                        child.just_happened = np.array([n.actA[i], n.actB[j]])
                        child.actA = n.actA[n.actA != n.actA[i]]
                        child.actB = n.actB[n.actB != n.actB[j]]
                        ### child.x = arithmetics.rand_f(1, len(child.actA) - 1)[0]
                        ### child.y = arithmetics.rand_f(1, len(child.actB) - 1)[0]
                        child_layer.append(+child)
            if len(child_layer) == 0:
                break
            else:
                parent_layer = child_layer
                child_layer = []

    def get_bfs_layers(self):
        bfs_layers = []
        parent_layer = [+self.root]
        child_layer = []
        while True:
            bfs_layers.append(parent_layer)
            for p in parent_layer:
                n = self.T.MEM[p]
                for i in range(len(n.actA)):
                    for j in range(len(n.actB)):
                        c = n.child[i, j]
                        child_layer.append(c)
            if len(child_layer) == 0:
                break
            else:
                parent_layer = child_layer
                child_layer = []
        return bfs_layers

    def payoff_init(self, bfs_layers):
        for p in bfs_layers[-1]:
            leaf = self.T.MEM[p]
            h = leaf.h()
            u, v = 0, 0
            for i in range(len(h)):
                p, q = h[i]
                if p > q:
                    u += self.upcard[i] * self.valueA[i]
                    v -= self.upcard[i] * self.valueB[i]
                elif p < q:
                    u -= self.upcard[i] * self.valueA[i]
                    v += self.upcard[i] * self.valueB[i]
            leaf.u, leaf.v = u, v
        for p in bfs_layers[-2]:
            trivial: tree.Node = self.T.MEM[p]
            leaf = trivial.get_child(0, 0)
            trivial.u, trivial.v = leaf.u, leaf.v

    def strategy_init(self, bfs_layers):
        nodes = []
        Ra = {}
        Rb = {}
        x = {}
        y = {}
        for l in bfs_layers[0:-2]:
            for p in l:
                nodes.append(p)
                n: tree.Node = self.T.MEM[p]
                Ra[p] = np.random.rand(*n.actA.shape) * 5
                Rb[p] = np.random.rand(*n.actB.shape) * 5
                x[p] = Ra[p] / Ra[p].sum()
                y[p] = Rb[p] / Rb[p].sum()
        return nodes, Ra, Rb, x, y

    def log_init(self, nodes: list, T: int, Ra: dict, Rb: dict, x: dict, y: dict):
        # RaT[ptr_node][act][time] EXCLUDE raT[ptr_node][act][time]!
        # This is a DIFFERENT bookkeeping/convention!
        RaT = {}
        RbT = {}
        xT = {}
        yT = {}
        for n in nodes:
            RaT[n] = np.zeros([*self.T.MEM[n].actA.shape, 1 + T])
            xT[n] = np.zeros([*self.T.MEM[n].actA.shape, 1 + T])
            RaT[n][:, 0] = Ra[n]
            xT[n][:, 0] = x[n]

            RbT[n] = np.zeros([*self.T.MEM[n].actB.shape, 1 + T])
            yT[n] = np.zeros([*self.T.MEM[n].actB.shape, 1 + T])
            RbT[n][:, 0] = Rb[n]
            yT[n][:, 0] = y[n]
        return T, RaT, RbT, xT, yT

    def payoff_refresh(self, bfs_layers, x, y):
        for l in bfs_layers[-3::-1]:
            for p in l:
                n = self.T.MEM[p]
                A = np.zeros([len(n.actA), len(n.actB)], dtype=float)
                B = np.zeros([len(n.actA), len(n.actB)], dtype=float)
                for i in range(len(n.actA)):
                    for j in range(len(n.actB)):
                        A[i, j] = self.T.MEM[n.child[i, j]].u
                        B[i, j] = self.T.MEM[n.child[i, j]].v
                n.u = (A @ x[p]) @ y[p]
                n.v = (B @ x[p]) @ y[p]

    def strategy_update(self, bfs_layers, Ra, Rb, x, y):
        for l in bfs_layers[-3::-1]:
            for p in l:
                # Evaluate
                n = self.T.MEM[p]
                A = np.zeros([len(n.actA), len(n.actB)], dtype=float)
                B = np.zeros([len(n.actA), len(n.actB)], dtype=float)
                for i in range(len(n.actA)):
                    for j in range(len(n.actB)):
                        A[i, j] = self.T.MEM[n.child[i, j]].u
                        B[i, j] = self.T.MEM[n.child[i, j]].v
                u = A @ y[p]
                v = B.T @ x[p]
                # Update
                ra = u - x[p].T @ u
                rb = v - y[p].T @ v
                ra[ra < 0] *= 0.5
                rb[rb < 0] *= 0.5
                Ra[p] += ra
                Rb[p] += rb
                Ra[p][Ra[p] < 0] = 0
                Rb[p][Rb[p] < 0] = 0
                # Output
                if Ra[p].sum():
                    x[p] = Ra[p] / Ra[p].sum()
                if Rb[p].sum():
                    y[p] = Rb[p] / Rb[p].sum()

    def log(self, nodes, t, Ra, Rb, x, y, RaT, RbT, xT, yT):
        for n in nodes:
            RaT[n][:, t] = Ra[n]
            RbT[n][:, t] = Rb[n]
            xT[n][:, t] = x[n]
            yT[n][:, t] = y[n]


def split_to_stackers(A: np.ndarray):
    M, N = A.shape
    y = []
    for i in range(M):
        y.append(A[i, :])
    return y


def init():
    RAND_SEED = 133484  # np.save("seed.npy", np.array([RAND_SEED])) # RAND_SEED: int = np.load("seed.npy")[0]
    numpy.random.seed(RAND_SEED)
    game = GoofSpiel()
    game.tree_init()  # np.save("goofspiel.npy", np.array([game])) # game: GoofSpiel = np.load("goofspiel.npy", allow_pickle=True)[0]
    bfs_layers = game.get_bfs_layers()
    game.payoff_init(bfs_layers)
    nodes, Ra, Rb, x, y = game.strategy_init(bfs_layers)
    T, RaT, RbT, xT, yT = game.log_init(nodes, 10000, Ra, Rb, x, y)
    for t in range(T):
        if t % 1000 == 0:
            print(t)
        game.payoff_refresh(bfs_layers, x, y)
        game.strategy_update(bfs_layers, Ra, Rb, x, y)
        game.log(nodes, t + 1, Ra, Rb, x, y, RaT, RbT, xT, yT)
    #---=---=---=---#---=---=---=---#---=---=---=---#---=---=---=---

    # visualization
    for p in Ra.keys():
        n = game.T.MEM[p]
        fig = plt.figure("history=" + str(n.h()), [8, 6], 96)
        fig.add_axes([0.05, 0.55, 0.4, 0.4])
        fig.add_axes([0.55, 0.55, 0.4, 0.4])
        fig.add_axes([0.05, 0.05, 0.4, 0.4])
        fig.add_axes([0.55, 0.05, 0.4, 0.4])
        fig.axes[0].stackplot(np.arange(1 + T), *split_to_stackers(xT[+n]), labels=n.actA)
        fig.axes[1].stackplot(np.arange(1 + T), *split_to_stackers(yT[+n]), labels=n.actB)
        fig.axes[2].stackplot(np.arange(1 + T), *split_to_stackers(RaT[+n]), labels=n.actA)
        fig.axes[3].stackplot(np.arange(1 + T), *split_to_stackers(RbT[+n]), labels=n.actB)
        for ax in fig.axes:
            ax.grid()
            ax.legend()
        fig.savefig("history=" + str(n.h()) + ".png")


if __name__ == "__main__":
    init()
